"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.AlgoTransaction = void 0;
const types_1 = require("../../types");
const algoUtils_1 = require("../../utils/algoUtils");
const constants_1 = require("../../utils/constants");
const errors_1 = require("../../utils/errors");
const managed_1 = require("../../utils/managed");
const utils_1 = require("../../utils/utils");
const TransactionBase_1 = require("../TransactionBase");
const web3 = require("web3");
/**
 * docs https://developer.algorand.org/docs/get-details/transactions/transactions/
 */
let AlgoTransaction = class AlgoTransaction extends TransactionBase_1.TransactionBase {
    get txid() {
        return this.data.txid;
    }
    get stdTxid() {
        return (0, algoUtils_1.base32ToHex)(this.data.txid);
    }
    get hash() {
        return this.txid;
    }
    get reference() {
        if (this.data.note) {
            return [(0, algoUtils_1.bytesToHex)(this.data.note)];
        }
        return [];
    }
    get stdPaymentReference() {
        let paymentReference = this.reference.length === 1 ? (0, utils_1.prefix0x)(this.reference[0]) : "";
        try {
            // try to parse out
            paymentReference = (0, utils_1.prefix0x)(web3.utils.hexToString((0, utils_1.prefix0x)(paymentReference)));
        }
        catch (e) {
            /* istanbul ignore next */
            // Should not happen
            return utils_1.ZERO_BYTES_32;
        }
        if (!(0, utils_1.isValidBytes32Hex)(paymentReference)) {
            paymentReference = utils_1.ZERO_BYTES_32;
        }
        return paymentReference;
    }
    get unixTimestamp() {
        return this.data.timestamp;
    }
    get sourceAddresses() {
        // for now mcc will only interpret the sender as the address that pays the fee (signs the transaction)
        return [(0, algoUtils_1.hexToBase32)(this.data.snd)];
    }
    get assetSourceAddresses() {
        if (this.type === "axfer" || this.type === "axfer_close") {
            // for token transfers send by clawback transaction (https://developer.algorand.org/docs/get-details/transactions/transactions/#asset-clawback-transaction)
            // and asset transfer transactions (https://developer.algorand.org/docs/get-details/transactions/transactions/#asset-transfer-transaction)
            if (this.data.asnd) {
                // for asset clawback tx it somehow gets addess w/out checksum
                return [(0, algoUtils_1.hexToBase32)((0, algoUtils_1.bufAddToCBufAdd)(this.data.asnd))];
            }
            // Other kinds of asset transactions (axfer and axfer_close ) use snd field as sender
            return [(0, algoUtils_1.hexToBase32)(this.data.snd)];
        }
        // Other kinds of transactions have nothing to do with assets
        return [];
    }
    get receivingAddresses() {
        const recAddresses = [];
        // for transactions of type pay
        if (this.type === "pay" || this.type === "pay_close") {
            if (this.data.rcv) {
                recAddresses.push((0, algoUtils_1.hexToBase32)(this.data.rcv));
            }
            // If address is closed all remaining founds will be transferred to this address
            if (this.data.close) {
                recAddresses.push((0, algoUtils_1.hexToBase32)(this.data.close));
            }
            return recAddresses;
        }
        return [];
    }
    get assetReceivingAddresses() {
        const recAddresses = [];
        // for token interaction transactions
        if (this.type === "axfer" || this.type === "axfer_close") {
            if (this.data.arcv) {
                recAddresses.push((0, algoUtils_1.hexToBase32)(this.data.arcv));
            }
            if (this.data.aclose) {
                recAddresses.push((0, algoUtils_1.hexToBase32)(this.data.aclose));
            }
            return recAddresses;
        }
        return [];
    }
    get fee() {
        return (0, utils_1.toBN)(this.data.fee || 0);
    }
    get spentAmounts() {
        // for transactions of type pay
        if (this.type === "pay" || "pay_close") {
            if (this.data.close) {
                // all assets that are not fee and received amount (amt) are transferred to close address
                throw new errors_1.mccError(errors_1.mccErrorCode.InvalidResponse, Error("Spend Amounts can't be extracted from transaction object on non-archival node"));
            }
            if (this.data.amt) {
                let amount = this.data.amt.toString();
                return [
                    {
                        address: (0, algoUtils_1.hexToBase32)(this.data.snd),
                        amount: this.fee.add((0, utils_1.toBN)(amount)),
                    },
                ];
            }
        }
        return [
            {
                // sender always pays the fee
                address: (0, algoUtils_1.hexToBase32)(this.data.snd),
                amount: this.fee,
            },
        ];
    }
    get assetSpentAmounts() {
        // for transactions of type axfer
        if (this.type === "axfer") {
            if (this.data.aamt) {
                let amount = this.data.aamt.toString();
                if (!(this.additionalData && this.additionalData.assetInfo)) {
                    throw new errors_1.mccError(errors_1.mccErrorCode.InvalidResponse, Error("call the getAllData before to make sure additional data full"));
                }
                let decimals = this.additionalData.assetInfo.params.decimals;
                if (typeof decimals === "bigint") {
                    decimals = Number(decimals);
                }
                return [
                    {
                        address: this.assetSourceAddresses[0],
                        amount: (0, utils_1.toBN)(amount),
                        elementaryUnits: web3.utils.toBN(Math.pow(10, decimals)),
                    },
                ];
            }
        }
        if (this.type === "axfer_close") {
            throw new errors_1.mccError(errors_1.mccErrorCode.InvalidResponse, Error("Asset Spent Amounts can't be extracted for axfer_close transaction from transaction object on non-archival node"));
        }
        return [];
        // For now we don't support spending / using non-native tokens
    }
    get receivedAmounts() {
        // for transactions of type pay
        if (this.type === "pay_close") {
            throw new errors_1.mccError(errors_1.mccErrorCode.InvalidResponse, Error("Received Amounts can't be extracted from transaction object on non-archival node for close remainder to transactions"));
        }
        if (this.data) {
            if (this.data.amt) {
                let amount = this.data.amt.toString();
                return [
                    {
                        address: this.receivingAddresses[0],
                        amount: (0, utils_1.toBN)(amount),
                    },
                ];
            }
        }
        return [];
    }
    get assetReceivedAmounts() {
        if (this.type === "axfer_close") {
            throw new errors_1.mccError(errors_1.mccErrorCode.InvalidResponse, Error("Received Amounts can't be extracted from transaction object on non-archival node for axfer close transactions"));
        }
        if (this.type === "axfer") {
            if (this.data.aamt) {
                let amount = this.data.aamt.toString();
                if (!(this.additionalData && this.additionalData.assetInfo)) {
                    throw new errors_1.mccError(errors_1.mccErrorCode.InvalidResponse, Error("call the getAllData before to make sure additional data full"));
                }
                let decimals = this.additionalData.assetInfo.params.decimals;
                if (typeof decimals === "bigint") {
                    decimals = Number(decimals);
                }
                return [
                    {
                        address: this.assetReceivingAddresses[0],
                        amount: (0, utils_1.toBN)(amount),
                        elementaryUnits: web3.utils.toBN(Math.pow(10, decimals)),
                    },
                ];
            }
        }
        return [];
    }
    get type() {
        if (this.data.type === "pay") {
            if (this.data.close) {
                return "pay_close";
            }
        }
        if (this.data.type === "axfer") {
            if (this.data.aclose) {
                return "axfer_close";
            }
        }
        return this.data.type;
    }
    get isNativePayment() {
        return this.type === "pay";
    }
    get currencyName() {
        if (this.type === "pay") {
            return constants_1.ALGO_NATIVE_TOKEN_NAME;
        }
        else if (this.type === "axfer" && this.data.xaid) {
            return this.data.xaid.toString();
        }
        return "";
    }
    get elementaryUnits() {
        return (0, utils_1.toBN)(constants_1.ALGO_MDU);
    }
    get successStatus() {
        // TODO research this further
        return types_1.TransactionSuccessStatus.SUCCESS;
    }
    async paymentSummary(client, inUtxo, utxo, makeFullPayment) {
        if (!this.isNativePayment) {
            if (this.type === "axfer") {
                if (client && makeFullPayment) {
                    await this.getTransactionAssetParams(client);
                    // TODO payment summary for algorand token transaction
                }
                else {
                    const ErrorMessage = "To get all information about token transactions makeFullPayment must be set to true and an instance of client must be provided";
                    throw new errors_1.mccError(errors_1.mccErrorCode.InvalidParameter, Error(ErrorMessage));
                }
            }
            return { isNativePayment: false };
        }
        return {
            isNativePayment: true,
            sourceAddress: this.sourceAddresses[0],
            receivingAddress: this.receivingAddresses[0],
            spentAmount: this.spentAmounts[0].amount,
            receivedAmount: this.receivedAmounts[0].amount,
            paymentReference: this.stdPaymentReference,
            oneToOne: true,
            isFull: true,
        };
    }
    async makeFull(client) {
        await this.getTransactionAssetParams(client);
    }
    //////////////////////////////////////////
    //// Algo transaction private methods ////
    //////////////////////////////////////////
    async getTransactionAssetParams(client) {
        if (this.data.xaid) {
            if (!this.additionalData) {
                this.additionalData = {};
            }
            if (!this.additionalData.assetInfo) {
                // @ts-ignore
                this.additionalData.assetInfo = await client.getAssetInfo(this.data.xaid);
            }
        }
    }
};
AlgoTransaction = __decorate([
    (0, managed_1.Managed)()
], AlgoTransaction);
exports.AlgoTransaction = AlgoTransaction;
//# sourceMappingURL=AlgoTransaction.js.map