import BN from "bn.js";
import { IUtxoGetTransactionRes } from "../..";
import { MccClient, MccUtxoClient, TransactionSuccessStatus } from "../../types";
import { IUtxoTransactionAdditionalData, IUtxoVinTransaction, IUtxoVinVoutsMapper, IUtxoVoutTransaction } from "../../types/utxoTypes";
import { AddressAmount, PaymentSummary, TransactionBase } from "../TransactionBase";
export declare type UtxoTransactionTypeOptions = "coinbase" | "payment" | "partial_payment" | "full_payment";
export declare class UtxoTransaction extends TransactionBase<IUtxoGetTransactionRes, IUtxoTransactionAdditionalData> {
    constructor(data: IUtxoGetTransactionRes, additionalData?: IUtxoTransactionAdditionalData);
    get txid(): string;
    get stdTxid(): string;
    get hash(): string;
    get reference(): string[];
    get stdPaymentReference(): string;
    get unixTimestamp(): number;
    get sourceAddresses(): (string | undefined)[];
    get assetSourceAddresses(): (string | undefined)[];
    get receivingAddresses(): (string | undefined)[];
    get assetReceivingAddresses(): (string | undefined)[];
    get fee(): BN;
    get spentAmounts(): AddressAmount[];
    get assetSpentAmounts(): AddressAmount[];
    get receivedAmounts(): AddressAmount[];
    get assetReceivedAmounts(): AddressAmount[];
    get type(): UtxoTransactionTypeOptions;
    get isNativePayment(): boolean;
    get currencyName(): string;
    get elementaryUnits(): BN;
    get successStatus(): TransactionSuccessStatus;
    paymentSummary(client?: MccClient, vinIndex?: number, voutIndex?: number, makeFullPayment?: boolean): Promise<PaymentSummary>;
    makeFull(client: MccClient): Promise<void>;
    /**
     * Asserts whether the vin index is in valid range. If not, exception is thrown.
     * @param vinIndex vin index
     */
    assertValidVinIndex(vinIndex?: number, canBeNull?: boolean): void;
    /**
     * Asserts whether the vout index is in valid range. If not, exception is thrown.
     * @param voutIndex vout index
     */
    assertValidVoutIndex(voutIndex?: number, canBeNull?: boolean): void;
    /**
     * Asserts whether vinvouts Mapper is full mapper and only has unique and matching indices
     */
    assertAdditionalData(): void;
    synchronizeAdditionalData(): void;
    /**
     * Extract vout on vout index. If vout index is not valid or data is corrupted, exception is thrown.
     * @param voutIndex vout index
     * @returns vout on vout index
     */
    extractVoutAt(voutIndex: number): IUtxoVoutTransaction;
    /**
     *
     * @param vinIndex vin index
     * @returns vin on vin index
     */
    extractVinAt(vinIndex: number): IUtxoVinTransaction;
    /**
     * Gets the vout corresponding to vin in the given vin index. If the vout is not fetched, the
     * input transaction is read and the corresponding vout is obtained
     * @param vinIndex vin index
     * @param client mcc client for fetching input the transactions
     * @returns input vout of the relevant utxo of the input transaction on the vin on the index `vinIndex`
     */
    private extractVinVoutAt;
    /**
     * Gets the vout corresponding to vin in the given vin index. If it is not stored in the additionalData,
     * it's fetched and updated using the client.
     * @param vinIndex vin index
     * @param client mcc client to fetch transaction data
     * @returns input vout of the relevant utxo of the input transaction on the vin on the index `vinIndex`
     */
    vinVoutAt(vinIndex: number, client?: MccUtxoClient): Promise<IUtxoVinVoutsMapper>;
    /**
     * Checks if the input on index `vinIndex` is fetched.
     * @param vinIndex vin index
     * @returns true if input on the index `vinIndex` is already fetched.
     */
    isSyncedVinIndex(vinIndex: number): boolean;
    /**
     * Fetches all the vin data from input transactions and makes the payment full_payment.
     * @param client mcc client to use for fetching input data
     */
    makeFullPayment(client: MccUtxoClient): Promise<void>;
    private processOutput;
}
//# sourceMappingURL=UtxoTransaction.d.ts.map